--[[
	Todo
		- confirm GetThreat() works with aloft.
		- Check if UpdateNameplateInfo() ever gets a nameplate GUID. RAID_TARGET_UPDATE and checking OnShow may be sufficient.
		- tidyPlatesRunning might not be needed. (frame.extended)
	Notes
		- hooking plates OnEnter requires the plates enable mouse. But doing that stops the mouseover ID being available. 
		I've tried disabing mouse once OnEnter fires but it's not guaranteed that the UPDATE_MOUSEOVER will then fire with the mobs moueoverID. This is more true when nameplates are clusterd ontop of each other.
		- TidyPlates sometimes removes the children from a plate.
		- TidyPlates seems to break my OnShow hook.
]]

local MAJOR, MINOR = "LibNameplate-1.0", 1
if not LibStub then error(MAJOR .. " requires LibStub.") return end

local lib = LibStub:NewLibrary(MAJOR, MINOR)
if not lib then return end

lib.callbacks = lib.callbacks or LibStub("CallbackHandler-1.0"):New(lib)
if not lib.callbacks then	error(MAJOR .. " CallbackHandler-1.0.") return end

-- Globals
local _G					= _G
local tostring				= tostring
local WorldFrame			= WorldFrame
local select				= select
local UnitExists			= UnitExists
local UnitGUID				= UnitGUID
local UnitName				= UnitName
local string_gsub			= string.gsub
local pairs					= pairs
local table_remove			= table.remove
local table_insert			= table.insert
local GetMouseFocus			= GetMouseFocus
local GetNumRaidMembers		= GetNumRaidMembers
local GetNumPartyMembers	= GetNumPartyMembers
local UnitIsUnit			= UnitIsUnit
local GetRaidTargetIndex	= GetRaidTargetIndex
local UnitInParty			= UnitInParty
local tonumber				= tonumber
local print					= print
local math_floor			= math.floor
local RAID_CLASS_COLORS		= RAID_CLASS_COLORS
local table_getn			= table.getn
local CreateFrame			= CreateFrame
local _ --localized underscore so FindGlobals doesn't bug me about it.

--Conts
local scanDelay		= 1 -- Scan for new nameplates every 1 seconds. We hook OnShow so this doesn't need to be done rapidly.
local updateDelay	= 1 -- Update nameplate info every 1 seconds. This checks if a nameplate has a raid icon then trys to find the GUID to that icon.

--these aren't really used. more to help me debug for now.
local aloftRunning			= false
local tidyPlatesRunning 	= false


local DEBUG = false
--[===[@debug@
DEBUG = true
--@end-debug@]===]

local function CmdHandler()
	DEBUG = not DEBUG
end
_G.SlashCmdList["LIBNAMEPLATEDEBUG"] = CmdHandler
_G.SLASH_LIBNAMEPLATEDEBUG1 = "/lnbug"
local function debugPrint(...)
	if DEBUG then
		print(...)
	end
end

local onShowHooks = {}
local onHideHooks = {}
local onUpdateHooks = {}

local regionNames = { "threatGlow", "healthBorder", "castBorder", "castNostop", "spellIcon", "highlightTexture", "nameText", "levelText", "dangerSkull", "raidIcon", "eliteIcon" }
local regions = {}
for i, name in pairs(regionNames) do 
	regions[name] = i
end

local barNames = { "healthBar", "castBar"}
local bars = {}
for i, name in pairs(barNames) do 
	bars[name] = i
end



--"Star", "Circle", "Diamond", "Triangle", "Moon", "Square", "Cross", "Skull"
--~ local raidIconTexture = "Interface\\TargetingFrame\\UI-RaidTargetingIcons"
--			= left							right
--			= ULx	ULy		LLx		LLy		URx		URy		LRx		LRy
-- star		= 0		0		0		0.25	0.25	0		0.25	0.25
-- circle	= 0.25	0		0.25	0.25	0.5		0		0.5		0.25
-- diamond	= 0.5	0		0.5		0.25	0.75	0		0.75	0.25
-- triangle	= 0.75	0		0.75	0.25	1		0		1		0.25
-- moon		= 0		0.25	0		0.5		0.25	0.25	0.25	0.5
-- square 	= 0.25	0.25	0.25	0.5		0.5		0.25	0.5		0.5
-- cross	= 0.5	0.25	0.5		0.5		0.75	0.25	0.75	0.5
-- skull	= 0.75	0.25	0.75	0.5		1		0.25	1		0.5
local raidIconTexCoord = {--from GetTexCoord. input is ULx and ULy (first 2 values).
	[0]		= {
		[0]		= 1, --star 
		[0.25]	= 5, --moon
	},
	[0.25]	= {
		[0]		= 2, --circle 
		[0.25]	= 6, --square
	},
	[0.5]	= {
		[0]		= 3, --star 
		[0.25]	= 7, --cross
	},
	[0.75]	= {
		[0]		= 4, --star 
		[0.25]	= 8, --skull
	},
}

lib.nameplates		= lib.nameplates or {}
lib.GUIDs			= lib.GUIDs or {}
lib.names			= lib.names or {}

local enableMouseFrames = {}

local plateGUIDs = {}

local colorToClass = {}
local function pctToInt(number) return math_floor((100*number) + 0.5) end
for classname, color in pairs(RAID_CLASS_COLORS) do
	colorToClass["C"..pctToInt(color.r)+pctToInt(color.g)+pctToInt(color.b)] = classname
end


--[[
	these always return nil
	frame:IsClampedToScreen(), frame:IsIgnoringDepth(), frame:IsMovable(), frame:IsToplevel(), frame:GetDepth(), frame:GetEffectiveDepth(), 

]]
local function IsNamePlateFrame(frame)
	if frame:GetName() then
--~ 		debugPrint("IsNamePlateFrame","No name!")
		return false
	end

	if frame:GetID() ~= 0 then
		return false
	end
	
	if frame:GetObjectType() ~= "Frame" then
--~ 		debugPrint("IsNamePlateFrame","Not a frame!")
		return false
	end
		
	if frame.extended or frame.aloftData then
		--Tidyplates = extended, Aloft = aloftData
		--They sometimes remove & replace the children so this needs to be checked first.
		return true
	end

	if frame:GetNumChildren() == 0 then
--~ 		debugPrint("IsNamePlateFrame","No children?!!", frame:GetName(), frame.extended and frame.extended.bars)
		return false
	end
		
	local children = {frame:GetChildren()}
	local sbFound = false
	for c = 1, table_getn(children) do
		if children[c]:GetObjectType() == "StatusBar" then
			sbFound = true
			break
		end
	end
	if sbFound == false then
--~ 		debugPrint("IsNamePlateFrame","No statusbar!")
		return false
	end

	
	return true
end



--[===[@debug@
function LC_TEST()--		/run LC_TEST()
	local c = {WorldFrame:GetChildren()}
	local frame, cFrame
	for i = 1, table_getn(c) do
		frame = c[i]
		if frame:IsShown() then
			--frame.extended
			debugPrint(IsNamePlateFrame(frame), "ID:"..tostring(frame:GetID()), "C:"..tostring(frame:GetNumChildren()), "R:"..tostring(frame:GetNumRegions()))
			
		end
	end
end
--@end-debug@]===]

local isOnScreen = {}

local lastChildren = 0
local function FindNameplates()
	local curChildren = WorldFrame:GetNumChildren()
	if curChildren ~= lastChildren then
		lastChildren = curChildren
		local frames = {WorldFrame:GetChildren()}
		local frame
		for i = 1, table_getn(frames) do
			frame = frames[i]
			if frame:IsShown() then
				if IsNamePlateFrame(frame) then
					if not lib.nameplates[frame] then
						lib:SetupNameplate(frame, true)
					elseif not isOnScreen[frame] then
						debugPrint("The hell? Did our OnShow hook fail?")
						onShowHooks[frame] = false
						lib:SetupNameplate(frame, true)
					end
				end
			end
		end
	end
--~ 	debugPrint("FindNameplates")
end

local function threatByColor( object , name)
	local redCan, greenCan, blueCan, alphaCan = object:GetVertexColor()
	local shown
	if not object:IsShown() then return "LOW" end
	if greenCan > .7 then return "MEDIUM" end
	if redCan > .7 then return "HIGH" end
end

local function reactionByColor(r, g, b, a)																											
	if r < .01 and b < .01 and g > .99 then return "FRIENDLY", "NPC" 
	elseif r < .01 and b > .99 and g < .01 then return "FRIENDLY", "PLAYER"
	elseif r > .99 and b < .01 and g > .99 then return "NEUTRAL", "NPC"
	else return "HOSTILE", "PLAYER" end
end

local function combatByColor(r, g, b, a) 
	return (r > .5 and g < .5)
end

function lib:GetName(frame)
	if frame.extended then
		return frame.extended.unit.name
	end

	--string_gsub returns 2 values. I just want to return 1.
	local name = string_gsub(select(regions.nameText, frame:GetRegions()):GetText(), "|cffffffff", "")
	return name
end

function lib:GetScale(frame)
	if frame.extended then
		frame = frame.extended
	end
	return frame:GetScale()
end

function lib:GetVisibleFrame(frame)
	if frame.extended then
		frame = frame.extended
	end
	return frame
end

function lib:GetLevel(frame)
	if frame.extended then
		return tonumber(frame.extended.unit.level or 0)
	elseif frame.aloftData then
		return tonumber(frame.aloftData.level or 0)
	end
	return tonumber(select(regions.levelText, frame:GetRegions()):GetText() or 0)
end

function lib:GetReaction(frame)
	if frame.extended then
		return frame.extended.unit.reaction
	elseif frame.aloftData then
		local r, g, b = frame.aloftData.originalHealthBarR, frame.aloftData.originalHealthBarG, frame.aloftData.originalHealthBarB
		return reactionByColor(r, g, b, 1)
	end
	return reactionByColor(select(bars.healthBar, frame:GetChildren()):GetStatusBarColor())
end

function lib:GetType(frame)
	if frame.extended then
		return frame.extended.unit.type
	end
	return select(2, reactionByColor(select(bars.healthBar, frame:GetChildren()):GetStatusBarColor()))
end

function lib:IsBoss(frame)
	if frame.extended then
		return frame.extended.unit.isBoss
	end
	return select(regions.dangerSkull, frame:GetRegions()):IsShown()
end

function lib:GetHealthMax(frame)
	if frame.extended then
		return tonumber(frame.extended.unit.healthmax or 0)
	end
	local _, max = select(bars.healthBar, frame:GetChildren()):GetMinMaxValues()
	return tonumber(max or 0)
end

--This will return nil if we're not in a PvP zone (like in citys)
function lib:GetClass(frame)
	if frame.extended then
		return frame.extended.unit.class
	end
	local red, green, blue = select(bars.healthBar, frame:GetChildren()):GetStatusBarColor()
	return colorToClass["C"..pctToInt(red)+pctToInt(green)+pctToInt(blue)]
end

function lib:IsElite(frame)
	if frame.extended then
		return frame.extended.unit.isElite
	elseif frame.aloftData then
		return frame.aloftData.stateIconRegion:IsShown()
	end
	return select(regions.eliteIcon, frame:GetRegions()):IsShown()
end

function lib:GetThreatSituation(frame)
	if frame.extended then
		return frame.extended.unit.threatSituation
	end
	return threatByColor(select(regions.threatGlow, frame:GetRegions()))
end

function lib:GetAlpha(frame)-- really?
	return frame:GetAlpha()
end

function lib:IsTarget(frame)
	return frame:IsShown() and frame:GetAlpha() == 1 and UnitExists("target")
end

function lib:GetHealth(frame)
	if frame.extended then
		return tonumber(frame.extended.unit.health or 0)
	end
	return tonumber(select(bars.healthBar, frame:GetChildren()):GetValue() or 0)
end

function lib:IsMouseover(frame)
	if frame.extended then
		return frame.extended.unit.isMouseover
	end
	return select(regions.highlightTexture, frame:GetRegions()):IsShown()
end

function lib:IsCasting(frame)
	if frame.extended then
		return frame.extended.unit.isCasting
	end
	return select(bars.castBar, frame:GetChildren()):IsShown()
end

function lib:IsInCombat(frame)
	if frame.extended then
		return frame.extended.unit.isInCombat
	end
	return combatByColor(select(regions.nameText, frame:GetRegions()):GetTextColor())
end

function lib:IsMarked(frame)
	if frame.extended then
		return frame.extended.unit.isMarked
	end
	return select(regions.raidIcon, frame:GetRegions()):IsShown()
end

function lib:GetRaidIcon(frame)
	local raidIcon
	if frame.extended then
		raidIcon = frame.extended.regions.raidIcon
	else
		raidIcon = select(regions.raidIcon, frame:GetRegions())
	end
	if raidIcon:IsShown() then
		local ULx,ULy = raidIcon:GetTexCoord() --,LLx,LLy,URx,URy,LRx,LRy
		if ULx and ULy then
			return raidIconTexCoord[ULx] and raidIconTexCoord[ULx][ULy] or 0
		end
	end
	return 0
end

function lib:GetGUID(frame)
	return plateGUIDs[frame]
end

-- ***
function lib:GetTargetNameplate()
	if self.targeted and self.targeted:IsShown() then
		return self.targeted
	end
end

function lib:GetNameplateByGUID(GUID)
	if self.GUIDs[GUID] and self.GUIDs[GUID]:IsShown() then
		return self.GUIDs[GUID]
	end
end


function lib:GetNameplateByName(name)
	if self.names[name] and self.names[name]:IsShown() then
		return self.names[name]
	end
end

function lib:GetNameplateByUnit(unitID)
	if UnitIsUnit(unitID, "target") then
		return self:GetTargetNameplate()
	end
	local GUID = UnitGUID(unitID)
	if self.GUIDs[GUID] and self.GUIDs[GUID]:IsShown() then
		return self.GUIDs[GUID]
	end
	local name = UnitName(unitID)
	if self.names[GUID] and self.names[GUID]:IsShown() then
		return self.names[GUID]
	end
end

local visibleFrame = {}
local function RecycleNameplate(frame)
	lib.callbacks:Fire("LibNameplate_RecycleNameplate", frame)
	
	if plateGUIDs[frame] then
		lib.GUIDs[plateGUIDs[frame]] = false
	end
	
	if visibleFrame[frame] then
		lib.names[visibleFrame[frame]] = false
	end
	
	plateGUIDs[frame] = false
	visibleFrame[frame] = false
end

local function FoundPlateGUID(frame, GUID, unitID)
	plateGUIDs[frame] = GUID
	
	lib.GUIDs[GUID] = frame
	lib.callbacks:Fire("LibNameplate_FoundGUID", frame, GUID, unitID)
end

local function GetMouseoverGUID(frame)
	local unitID = "mouseover"
	if UnitExists(unitID) then
		FoundPlateGUID(frame, UnitGUID(unitID), unitID)
	end
end

local function FindPlateWithRaidIcon(iconNum)
	for frame in pairs(lib.nameplates) do 
		if lib:IsMarked(frame) and lib:GetRaidIcon(frame) == iconNum then
			return frame
		end
	end
	return nil
end

local function FindGUIDByRaidIcon(frame, raidNum, from)
	local group, num = "", 0
	if GetNumRaidMembers() > 1 then
		group, num = "raid", GetNumRaidMembers()
	elseif GetNumPartyMembers() > 0 then
		group, num = "party", GetNumPartyMembers()
	end
	
	local unitID, targetID, targetIndex
	for i = 1, num do
		unitID = "party"..i;
		targetID = unitID.."target"
		if not UnitIsUnit("target", targetID) then
			targetIndex = GetRaidTargetIndex(targetID)
			if targetIndex and targetIndex == raidNum then
				debugPrint("FindGUIDByRaidIcon", from, "Icon:"..tostring(raidNum), "unitID:"..tostring(targetID), "GUID:"..tostring(UnitGUID(targetID)))
				FoundPlateGUID(frame, UnitGUID(targetID), targetID)
				return
			end
		end
	end
end

local function UpdateNameplateInfo(frame)
	if not plateGUIDs[frame] then
		if lib:IsMouseover(frame) then
			GetMouseoverGUID(frame)
		elseif lib:IsMarked(frame) then
			local raidNum = lib:GetRaidIcon(frame)
			if raidNum and raidNum > 0 then
				FindGUIDByRaidIcon(frame, raidNum, "UpdateNameplateInfo")
			end
		end
	end

--~ 	debugPrint("Updated nameplate", lib:GetName(frame))
	frame.lnpLastUpdate = 0
end

local function ourOnShow(self, ...)
	lib:SetupNameplate(self)
end

local function ourOnHide(self, ...)
	isOnScreen[self] = false
	RecycleNameplate(self)
end

local callbackPlateChange = {}
local watchForPlateChange = false --Only watch for visible plate change if a addon has registered the callback.

function lib.callbacks:OnUsed(target, eventname)
	if eventname == "LibNameplate_VisiblePlateChange" then
		table_insert(callbackPlateChange, target)
		watchForPlateChange = true
		
--~ 		debugPrint("OnUsed", target, eventname, tidyPlatesRunning)
	end
end

function lib.callbacks:OnUnused(target, eventname)
	if eventname == "LibNameplate_VisiblePlateChange" then
		for i=table_getn(callbackPlateChange),1, -1  do 
			if callbackPlateChange[i] == target then
				table_remove(callbackPlateChange, i)
				break
			end
		end
		if table_getn(callbackPlateChange) == 0 then
			watchForPlateChange = false
		end
--~ 		debugPrint("OnUnused", target, eventname, watchForPlateChange, tidyPlatesRunning)
	end
end


local function ourOnUpdate(self, elapsed, ...)
	if self.lnpCheckForTarget then --Check on the first OnUpdate after the frame's shown.
		self.lnpCheckForTarget = false
		if not plateGUIDs[self] and self:IsShown() and ((self:GetAlpha() == 1) and UnitExists("target")) then
			FoundPlateGUID(self, UnitGUID("target"), "target")
		end
	end
	
	--[[]]
	self.lnpLastUpdate = (self.lnpLastUpdate or 0) + elapsed
	if self.lnpLastUpdate > updateDelay then
		UpdateNameplateInfo(self)
		
		if watchForPlateChange and not visibleFrame[self] then
			if self.extended then
				lib.callbacks:Fire("LibNameplate_VisiblePlateChange", self, self.extended)
				visibleFrame[self] = true
			end
		end

	end
end


function lib:SetupNameplate(frame)
	isOnScreen[frame] = true

	if frame:HasScript("OnHide") and not onHideHooks[frame] then
		onHideHooks[frame] = true
		frame:HookScript("OnHide", ourOnHide)
	end
	
	if frame:HasScript("OnShow") and not onShowHooks[frame] then
		onShowHooks[frame] = true
		frame:HookScript("OnShow", ourOnShow)
	end

	if frame:HasScript("OnUpdate") and not onUpdateHooks[frame] then
		onUpdateHooks[frame] = true
		frame:HookScript("OnUpdate", ourOnUpdate)
	end
	
	--[===[@debug@
	if frame:GetNumChildren() == 0 and tidyPlatesRunning == false then
		debugPrint("SetupNameplate 1", "Where'd all the children go?", frame.extended, frame.aloftData)
	end
	
	if frame:GetNumRegions() == 0 and tidyPlatesRunning == false then
		debugPrint("SetupNameplate 1", "Where'd all the regions go?", frame.extended, frame.aloftData)
	end
	--@end-debug@]===]
	
	local plateName = self:GetName(frame)
	lib.nameplates[frame] = plateName
	lib.names[plateName] = frame
	
	if watchForPlateChange and frame.extended then
		visibleFrame[frame] = true--We already know the visible frame. Addon can call GetVisibleFrame() to get it.
	end
	
	lib.callbacks:Fire("LibNameplate_NewNameplate", frame)

	frame.lnpCheckForTarget = true
	UpdateNameplateInfo(frame)
end--	/lnbug

local function CheckForTargetGUID()
	local unitID = "target"
	local GUID
	for frame in pairs(lib.nameplates) do 
		if lib:IsTarget(frame) then
			lib.targeted = frame
			if not plateGUIDs[frame] then
				FoundPlateGUID(frame, UnitGUID(unitID), unitID)
			end
			lib.callbacks:Fire("LibNameplate_TargetNameplate", frame)
			return
		end
	end
--~ 	debugPrint("Failed to find target GUID")
end

local function OnUpdate(self, elapsed)
	if self.checkTarget then
		self.checkTarget = false
		CheckForTargetGUID()
	end

	self.lastUpdate = self.lastUpdate + elapsed
	if self.lastUpdate >= scanDelay then
		self.lastUpdate = 0
		FindNameplates()
	end
end

local function OnEvent(self, event, ...)
	if event == "UPDATE_MOUSEOVER_UNIT" then
		if GetMouseFocus():GetName() == "WorldFrame" then
			for frame in pairs(lib.nameplates) do 
				if frame:IsShown() and lib:IsMouseover(frame) then
					if not plateGUIDs[frame] then
						GetMouseoverGUID(frame)
					end
					lib.callbacks:Fire("LibNameplate_MouseoverNameplate", frame)
					return
				end
			end
		end
		
	elseif event == "PLAYER_TARGET_CHANGED" then
		if UnitExists("target") then
			self.checkTarget = true--Check for target nameplate on next OnUpdate. We can't do it now.
		else
			lib.targeted = nil
		end
		
	elseif event == "UNIT_TARGET" then
		local unitID = ...
		local targetID = unitID.."target"
		if not UnitIsUnit("player", unitID) then
			local iconNum = GetRaidTargetIndex(targetID)
			if iconNum and iconNum > 0 then
				local foundPlate = FindPlateWithRaidIcon(iconNum)
				if foundPlate and not plateGUIDs[foundPlate] then
--~ 					debugPrint(event, "Found raid icon on ", UnitName(unitID), "'s target", UnitName(targetID), "icon:"..tostring(iconNum))
					FoundPlateGUID(foundPlate, UnitGUID(targetID), targetID)
				end
			end
		end
		
	elseif event == "RAID_TARGET_UPDATE" then
		for frame in pairs(lib.nameplates) do 
			if frame:IsShown() and not plateGUIDs[frame] and lib:IsMarked(frame) then
				local raidNum = lib:GetRaidIcon(frame)
				if raidNum and raidNum > 0 then
					FindGUIDByRaidIcon(frame, raidNum, event)
				end
			end
		end
	elseif event == "PLAYER_ENTERING_WORLD" then
--~ 		debugPrint(event, MAJOR, _G.TidyPlates, _G.Aloft)
		if _G.TidyPlates then
			tidyPlatesRunning = true
		elseif _G.Aloft then
			aloftRunning = true
		end
		
	end
end


if not lib.frame then
	lib.frame=CreateFrame("Frame")
end

lib.frame.lastUpdate = 0
lib.frame.checkTarget = false
lib.frame:SetScript("OnUpdate", OnUpdate)
lib.frame:SetScript("OnEvent", OnEvent)
lib.frame:RegisterEvent('UPDATE_MOUSEOVER_UNIT')
lib.frame:RegisterEvent('PLAYER_TARGET_CHANGED')
lib.frame:RegisterEvent('UNIT_TARGET')
lib.frame:RegisterEvent('RAID_TARGET_UPDATE')
lib.frame:RegisterEvent('PLAYER_ENTERING_WORLD')